﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Text;
using EnvelopeCreatorProject.Exceptions;
using EnvelopeCreatorProject.Interfaces;
using log4net;
using EnvelopeCreatorProject.Models.Builders;
using Newtonsoft.Json;

namespace EnvelopeCreatorProject.Models.Signers
{
    public abstract class OAuthSigner : ISigner
    {
        private const int OAUTH_ERROR_CODE = 3;
        private readonly log4net.ILog logger = LogManager.GetLogger("Signer");
        public string AuthToken { get; set; }
        public string BaseURL { get; set; }
        internal abstract void SetHeaderCollection(HttpWebRequest request);


        protected virtual bool ValidateAuthInfo(){
            return !String.IsNullOrEmpty(AuthToken) && !String.IsNullOrEmpty(BaseURL);
        }

        public EnvelopeResult GenerateEnvelope(EnvelopeData EData)
        {
            if (!this.ValidateAuthInfo())
                throw new OAuthTokenAuthorizationException(OAUTH_ERROR_CODE, "Invalid Docusign Auth Info. Please, make sure you have all OAuthTokens required for this transaction.");
    

            string jsonBody = "";
            try
            {
                jsonBody = GetRequestBody(EData);
            }
            catch (ArgumentNullException ex)
            {
                return new EnvelopeResult()
                {
                    EnvelopeResponse = null,
                    ErrorCode = 1,
                    Message = "Invalid Parameter. " + ex.ParamName
                };
            }

            string data = this.CreateRESTRequestData(jsonBody);
            byte[] requestBodyEnd = Encoding.UTF8.GetBytes("\r\n--DOCUSIGNENVELOPE--");
            byte[] recipientData = Encoding.UTF8.GetBytes(data);
            string result = "";
            string url = BaseURL + "/envelopes";
            int filesLength = 0;
            int count = 0;
            foreach (var document in EData.Documents)
            {
                StringBuilder requestData = new StringBuilder();
                AppendToRequest(document, requestData, count);
                byte[] pdfBytes = document.PDF;
                filesLength += pdfBytes.Length + Encoding.UTF8.GetBytes(requestData.ToString()).Length;
                count++;
            }
            try
            {
                HttpWebRequest request = (HttpWebRequest)HttpWebRequest.Create(url);
                Stream requestDataStream = null;
                BuildRequest(request, requestDataStream, EData.Documents, requestBodyEnd, recipientData, url, filesLength);               

                HttpWebResponse response = null;
                request.Timeout = 60000;
                response = (HttpWebResponse)request.GetResponse();
                Stream responseDataStream = response.GetResponseStream();
                int length = Convert.ToInt32(response.Headers["Content-Length"]);
                BinaryReader reader = new BinaryReader(responseDataStream);
                recipientData = reader.ReadBytes(length);
                result = UTF8Encoding.UTF8.GetString(recipientData);
            }
            catch (WebException ex)
            {
                var responseMessage = "";
                try
                {
                    var fullResponse = new StreamReader(((WebException) ex).Response.GetResponseStream()).ReadToEnd();
                    var dic = JsonConvert.DeserializeObject<Dictionary<string, string>>(fullResponse);
                    responseMessage = "Docusign's response: " + dic["message"];

                    logger.Error(string.Format("Create Envelope - AuthToken: {0}. Error: {1}", AuthToken, fullResponse));
                }
                catch (Exception)
                {
                    //Could not read response message. Leave empty.
                }

                return new EnvelopeResult()
                {
                    EnvelopeResponse = null,
                    ErrorCode = 1,
                    Message = string.Format("There was an error while calling Docusign. Please, contact an administrator. {0}", responseMessage) 
                };
            }

            return new EnvelopeResult() { 
                EnvelopeResponse = result,
                ErrorCode = 0,
                Message = "Envelope created successfully"
            };
        }

        private string CreateRESTRequestData(string jsonBody)
        {
            StringBuilder requestData = new StringBuilder();

            //Add first part of the data: jsonBody with recipient data
            requestData.AppendLine("--DOCUSIGNENVELOPE");
            requestData.AppendLine("Content-Type: application/json");
            requestData.AppendLine("Content-Disposition: form-data");
            requestData.AppendLine();
            requestData.AppendLine(jsonBody);

            return requestData.ToString();
        }

        private void AppendToRequest(DocusignDocument document, StringBuilder requestData, int count)
        {
            if (count != 0)
            {
                requestData.AppendLine("");
            }
            requestData.AppendLine("--DOCUSIGNENVELOPE");
            requestData.AppendLine("Content-Type: application/pdf");

            var id = document.ID.ToString();
            requestData.AppendLine("Content-Disposition: file; filename=" + document.FileName + ";documentId=" + id + "");
            requestData.AppendLine();
        }

        private void BuildRequest(HttpWebRequest request,
                                    Stream requestDataStream,
                                    List<DocusignDocument> documents,
                                    byte[] requestBodyEnd,
                                    byte[] recipientData,
                                    string url,
                                    int filesLength)
        {
            request.Method = "POST";
            request.ContentType = "multipart/form-data; boundary=DOCUSIGNENVELOPE";
            request.ContentLength = recipientData.Length + requestBodyEnd.Length + filesLength;
            request.Credentials = CredentialCache.DefaultCredentials;
            SetHeaderCollection(request);
            request.Accept = "application/json";
            requestDataStream = request.GetRequestStream();
            requestDataStream.Write(recipientData, 0, recipientData.Length);
            int count = 0;
            foreach (var document in documents)
            {
                StringBuilder requestData = new StringBuilder();
                AppendToRequest(document, requestData, count);
                byte[] pdfBytes = document.PDF;
                byte[] header = Encoding.UTF8.GetBytes(requestData.ToString());
                requestDataStream.Write(header, 0, header.Length);
                requestDataStream.Write(pdfBytes, 0, pdfBytes.Length);
                count++;
            }
            requestDataStream.Write(requestBodyEnd, 0, requestBodyEnd.Length);
            requestDataStream.Close();
        }

        private string GetRequestBody(EnvelopeData EData)
        {
            var envelopeBuilders = new List<IEnvelopeBuilder>(){
               {new EnvelopeStatusBuilder()},
               {new EnvelopeMailBuilder(EData.MailBody, EData.MailSubject)},
               {new EnvelopeSettingsBuilder(EData.EnvelopeSettings)},
               {new EnvelopeDocumentBuilder(EData.Documents)},
               {new EnvelopeRecipientBuilder(EData.Recipients, EData.Tabs, EData.DateSignedFields, EData.Host)}
            };
 
            // Start the JSON body for the REST request to Docusign
            StringBuilder jsonBody = new StringBuilder();
            jsonBody.AppendLine("{");
            envelopeBuilders.ForEach(b => b.BuildTo(jsonBody));
            jsonBody.Append("}");

            return jsonBody.ToString();
        }

       
    }
}
